# -*- coding: utf-8 -*-
# This file is part of the investment-module from m-ds for Tryton.
# The COPYRIGHT file at the top level of this repository contains the
# full copyright notices and license terms.

from trytond.model import ModelView, ModelSQL, fields, Index
from trytond.pyson import Eval, If, Bool
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.report import Report
from decimal import Decimal
from sql.functions import Round, CurrentDate
from sql.conditionals import NullIf, Coalesce, Greatest


class LineContext(ModelView):
    'Watchlist Context'
    __name__ = 'investment_watchlist.line.context'

    watchlist = fields.Many2One(
        string='Watchlist', readonly=True,
        model_name='investment_watchlist.list',
        states={'invisible': True})

# end LineContext


class WatchlistEntry(ModelSQL, ModelView):
    'Watchlist Entry'
    __name__ = 'investment_watchlist.line'

    watchlist = fields.Many2One(
        string='Watchlist', required=True,
        model_name='investment_watchlist.list', ondelete='CASCADE',
        states={
            'readonly': ~Bool(Eval('watchlist'))})
    asset = fields.Many2One(
        string='Asset', required=True,
        model_name='investment.asset', ondelete='RESTRICT')
    asset_currency = fields.Function(fields.Many2One(
        string='Asset Currency',
        model_name='currency.currency', readonly=True), 'get_entry_data')
    currency = fields.Function(fields.Many2One(
        string='Currency',
        model_name='currency.currency', readonly=True), 'get_entry_data')
    currency_digits = fields.Function(fields.Integer(
        string='Currency Digits',
        readonly=True), 'get_entry_data')
    price_digits = fields.Function(fields.Integer(
        string='Price Digits', readonly=True), 'get_entry_data')
    uom = fields.Function(fields.Many2One(
        string='UOM', readonly=True,
        model_name='product.uom'), 'get_entry_data')
    uom_digits = fields.Function(fields.Integer(
        string='Unit Digits', readonly=True), 'get_entry_data')
    asset_digits = fields.Function(fields.Integer(
        string='Asset Digits', readonly=True), 'get_entry_data')
    amount_digits = fields.Function(fields.Integer(
        string='Purchase Digits', readonly=True), 'get_entry_data')
    symbol = fields.Function(fields.Char(
        string='UOM', readonly=True), 'get_entry_data')

    # observe data
    date = fields.Date(
        string='Date', required=True, help='Date of price observation')
    price = fields.Numeric(
        string='Rate', required=True,
        help='Price of the asset at the time of acquisition.',
        digits=(16, Eval('price_digits', 4)),
        depends=['price_digits'])
    quantity = fields.Numeric(
        string='Quantity', required=True,
        help='Number of pieces in the depot.',
        digits=(16, Eval('uom_digits', 4)),
        depends=['uom_digits'])
    expenses = fields.Numeric(
        string='Expenses', required=True,
        help='Sum of all fees during purchase.',
        digits=(16, Eval('currency_digits', 2)),
        depends=['currency_digits'])
    amount = fields.Function(fields.Numeric(
        string='Purchase price',
        help='Value of the purchase including fees.',
        readonly=True, digits=(16, Eval('amount_digits', 4)),
        depends=['amount_digits']), 'get_line_amounts')

    # current values from asset
    asset_date = fields.Function(fields.Date(
        string='Date', readonly=True,
        help='Date of current rate'), 'get_line_amounts')
    asset_rate = fields.Function(fields.Numeric(
        string='Rate', readonly=True, help='Current rate',
        depends=['asset_digits'],
        digits=(16, Eval('asset_digits', 4))), 'get_line_amounts')
    asset_rate_cur = fields.Function(fields.Numeric(
        string='Rate', readonly=True,
        help='Current rate in watchlist-currency',
        states={
            'invisible': Eval('currency', -1) == Eval('asset_currency', -1),
        }, depends=['asset_digits', 'currency', 'asset_currency'],
        digits=(16, Eval('asset_digits', 4))), 'get_line_amounts')
    asset_amount = fields.Function(fields.Numeric(
        string='act.Value', help='Current value of the aquisition.',
        readonly=True, depends=['currency_digits'],
        digits=(16, Eval('currency_digits', 4))), 'get_line_amounts')

    # performance
    diff_amount = fields.Function(fields.Numeric(
        string='Difference',
        help='Difference between acquisition value and current value',
        readonly=True, digits=(16, Eval('currency_digits', 2)),
        depends=['currency_digits']), 'get_line_amounts')
    diff_percent = fields.Function(fields.Numeric(
        string='Percent', help='percentage performance since acquisition',
        readonly=True, digits=(16, Eval('currency_digits', 2)),
        depends=['currency_digits']), 'get_line_amounts')

    @classmethod
    def __setup__(cls):
        super(WatchlistEntry, cls).__setup__()
        cls._order.insert(0, ('asset.name', 'ASC'))
        cls._order.insert(0, ('asset.date', 'DESC'))
        t = cls.__table__()
        cls._sql_indexes.update({
            Index(
                t,
                (t.asset, Index.Equality())),
            })

    @classmethod
    def view_attributes(cls):
        return super(WatchlistEntry, cls).view_attributes() + [
            ('/tree', 'visual',
                If(Eval('diff_percent', 0) < 0, 'danger',
                    If(Eval('diff_percent', 0) > 0, 'success', ''))),
            ]

    @classmethod
    def get_currency_rate_sql(cls):
        """ table of currency-rates at date
        """
        pool = Pool()
        Currency = pool.get('currency.currency')
        CurRate = pool.get('currency.currency.rate')
        tab_cur = Currency.__table__()
        tab_cur_rate = CurRate.__table__()
        context = Transaction().context

        query_date = context.get('qdate', CurrentDate())
        tab_cur_rate = tab_cur.join(
                tab_cur_rate,
                condition=(tab_cur.id == tab_cur_rate.currency) &
                (tab_cur_rate.date <= query_date),
                type_='LEFT OUTER',
            ).select(
                tab_cur.id,
                tab_cur.digits,
                Coalesce(NullIf(tab_cur_rate.rate, 0.0), 1.0).as_('rate'),
                distinct_on=[tab_cur.id],
                order_by=[tab_cur.id, tab_cur_rate.date.desc])
        return tab_cur_rate

    @classmethod
    def get_line_amounts_sql(cls):
        """ get amounts of line, converted to wachtlist-currency
        """
        pool = Pool()
        Asset = pool.get('investment.asset')
        WatchList = pool.get('investment_watchlist.list')

        # currencies: from, to
        tab_watchlist2 = WatchList.__table__()
        tab_entry2 = cls.__table__()
        tab_asset2 = Asset.__table__()
        tab_currate_from = cls.get_currency_rate_sql()
        tab_currate_to = cls.get_currency_rate_sql()

        tab_currency = tab_watchlist2.join(
                tab_entry2,
                condition=tab_watchlist2.id == tab_entry2.watchlist,
            ).join(
                tab_asset2,
                condition=tab_entry2.asset == tab_asset2.id,
            ).join(
                tab_currate_from,
                condition=tab_asset2.currency == tab_currate_from.id,
            ).join(
                tab_currate_to,
                condition=tab_watchlist2.currency == tab_currate_to.id,
            ).select(
                tab_entry2.id,
                (tab_currate_to.rate /
                    tab_currate_from.rate).as_('to_from_factor'),
                tab_currate_to.digits.as_('to_digits'))

        tab_asset = Asset.__table__()
        tab_watchlist = WatchList.__table__()
        tab_entry = cls.__table__()
        (tab_rate, tab2) = Asset.get_rate_data_sql()

        tab_amounts = tab_entry.join(
                tab_rate,
                condition=tab_entry.asset == tab_rate.id,
            ).join(
                tab_watchlist,
                condition=tab_entry.watchlist == tab_watchlist.id,
            ).join(
                tab_asset,
                condition=tab_entry.asset == tab_asset.id,
            ).join(
                tab_currency,
                condition=tab_entry.id == tab_currency.id,
            ).select(
                tab_entry.id,
                tab_rate.date.as_('asset_date'),
                # observe amount @ asset-currency
                Round(
                    tab_entry.quantity * tab_entry.price,
                    Greatest(tab_watchlist.uom_digits, tab_currency.to_digits)
                    ).as_('obs_amount'),
                # observe amount @ watchlist-currency
                Round(
                    tab_entry.quantity * tab_entry.price *
                    tab_currency.to_from_factor,
                    Greatest(tab_watchlist.uom_digits, tab_currency.to_digits)
                    ).as_('obs_amount_cur'),
                # current rate @ asset-currency
                Round(
                    tab_rate.rate,
                    tab_asset.currency_digits).as_('asset_rate'),
                Round(
                    tab_rate.rate * tab_currency.to_from_factor,
                    tab_asset.currency_digits).as_('asset_rate_cur'),
                # current value of asset @ watchlist-currency
                Round(
                    tab_entry.quantity * tab_rate.rate,
                    tab_currency.to_digits).as_('asset_amount'),
                Round(
                    tab_entry.quantity * tab_rate.rate *
                        tab_currency.to_from_factor,
                    tab_currency.to_digits).as_('asset_amount_cur'),
                tab_entry.expenses,
                tab_currency.to_digits,
                tab_entry.watchlist)
        return (tab_amounts, tab_entry)

    @classmethod
    def get_line_amounts(cls, entries, names):
        """ get digits and units
        """
        cursor = Transaction().connection.cursor()

        result = {x: {y.id: None for y in entries} for x in names}

        def quant_value(value, digits):
            """ get digits string
            """
            quant = Decimal('1')
            if digits > 0:
                quant = Decimal('%s' % (1/10**digits))
            if value is None:
                return None
            return value.quantize(quant)

        if entries:
            (query, tab_entry) = cls.get_line_amounts_sql()
            query.where = tab_entry.id.in_([x.id for x in entries])
            cursor.execute(*query)
            records = cursor.fetchall()

            for record in records:
                values = {
                    'asset_date': record[1],
                    'amount': record[3] + record[8],
                    'asset_rate': record[4],
                    'asset_rate_cur': record[5],
                    'asset_amount': record[7],
                    'diff_amount': quant_value(
                        record[7] - (record[3] + record[8]), record[9])
                    if (record[7] - record[3]) != Decimal('0.0')
                    else Decimal('0.0'),
                    'diff_percent': quant_value(
                        Decimal('100.0') * record[7] /
                        (record[3] + record[8]) - Decimal('100.0')
                        if (record[3] + record[8]) != Decimal('0.0')
                        else Decimal('0.0'), record[9])
                    if (record[7] - record[3]) != Decimal('0.0')
                    else Decimal('0.0'),
                    }

                for name in names:
                    result[name][record[0]] = values[name]
        return result

    @classmethod
    def get_entry_data_sql(cls):
        """ sql for line-values
        """
        pool = Pool()
        Asset = pool.get('investment.asset')
        WatchList = pool.get('investment_watchlist.list')
        Currency = pool.get('currency.currency')

        tab_watchlist = WatchList.__table__()
        tab_entry = cls.__table__()
        tab_asset = Asset.__table__()
        tab_cur = Currency.__table__()
        (tab_symbol, tab2) = Asset.get_name_symbol_sql()

        query = tab_watchlist.join(
                tab_entry,
                condition=tab_watchlist.id == tab_entry.watchlist,
            ).join(
                tab_cur,
                condition=tab_watchlist.currency == tab_cur.id,
            ).join(
                tab_asset,
                condition=tab_entry.asset == tab_asset.id,
            ).join(
                tab_symbol,
                condition=tab_entry.asset == tab_symbol.id,
            ).select(
                tab_entry.id,
                tab_watchlist.currency.as_('currency'),     # watchlist-currency
                tab_cur.digits.as_('currency_digits'),
                tab_watchlist.price_digits,                 # digits set in
                tab_watchlist.uom_digits,                   # watchlist
                tab_asset.currency_digits.as_('asset_digits'),
                tab_asset.currency.as_('asset_currency'),
                Greatest(
                    tab_watchlist.uom_digits,
                    tab_cur.digits).as_('amount_digits'),
                tab_asset.uom,
                tab_symbol.symbol,
                tab_entry.watchlist)
        return (query, tab_entry)

    @classmethod
    def get_entry_data(cls, entries, names):
        """ get digits and units
        """
        (query, tab_entry) = cls.get_entry_data_sql()
        cursor = Transaction().connection.cursor()

        result = {x: {y.id: None for y in entries} for x in names}
        if not entries:
            return result

        query.where = tab_entry.id.in_([x.id for x in entries])

        cursor.execute(*query)
        records = cursor.fetchall()

        for record in records:
            values = {
                'currency': record[1],
                'currency_digits': record[2],
                'price_digits': record[3],
                'uom_digits': record[4],
                'asset_digits': record[5],
                'asset_currency': record[6],
                'amount_digits': record[7],
                'uom': record[8],
                'symbol': record[9],
                }

            for name in names:
                result[name][record[0]] = values[name]
        return result

    def get_rec_name(self, name):
        """ record name
        """
        return ' '.join([
            self.asset.product.rec_name,
            '|',
            Report.format_number(
                self.asset_amount, lang=None,
                digits=self.currency_digits or 2)
            if self.asset_amount is not None else '-',
            self.currency.symbol,
            '(%(proz)s %%, %(diff)s %(unit)s)' % {
                'proz': Report.format_number(
                    self.diff_percent, lang=None,
                    digits=self.currency_digits or 2)
                if self.diff_percent is not None else '-',
                'diff': Report.format_number(
                    self.diff_amount, lang=None,
                    digits=self.currency_digits or 2)
                if self.diff_amount is not None else '-',
                'unit': self.currency.symbol},
            '|',
            Report.format_date(self.asset_date)
            if self.asset_date is not None else '-'
            ])

    field_price = [
        'price', 'quantity', 'amount', 'expenses', 'price_digits',
        'currency', 'asset_currency']

    @fields.depends(*field_price)
    def on_change_quantity(self):
        """ calculate amount
        """
        self.on_change_price()

    @fields.depends(*field_price)
    def on_change_expenses(self):
        """ calculate amount
        """
        self.on_change_price()

    @fields.depends(*field_price)
    def on_change_price(self):
        """ calculate amount
        """
        Currency = Pool().get('currency.currency')

        digits = self.price_digits if self.price_digits is not None else 4
        exp = Decimal(Decimal(1) / 10 ** digits)

        if self.asset_currency and self.currency:
            self.amount = ((Currency.compute(
                self.asset_currency, self.price, self.currency,
                False) if self.price is not None else Decimal('0.0')) * (
                self.quantity
                if self.quantity is not None else Decimal('0.0')) - (
                self.expenses
                if self.expenses is not None else Decimal('0.0'))
                ).quantize(exp)

    @fields.depends(
        'asset', 'price', 'quantity', 'amount', 'expenses',
        'asset_date', 'asset_rate', 'asset_amount', 'diff_amount')
    def on_change_asset(self):
        """ update price+date
        """
        if self.asset:
            if self.asset_rate:
                self.asset_date = self.asset.date
                self.asset_rate = self.asset.rate
            if self.price is None:
                self.price = self.asset.rate
                self.date = self.asset.date
            if self.quantity is None:
                self.quantity = Decimal('1.0')
            if self.expenses is None:
                self.expenses = Decimal('0.0')
        else:
            self.price = None
            self.asset_rate = None
            self.date = None
            self.amount = None
            self.asset_date = None
            self.quantity = None
            self.expenses = None
            self.asset_amount = None
            self.diff_amount = None

    @classmethod
    def default_watchlist(cls):
        """ watchlist from context
        """
        context = Transaction().context
        return context.get('watchlist', None)

    @classmethod
    def default_expenses(cls):
        """ default 0
        """
        return Decimal('0.0')

    @classmethod
    def default_quantity(cls):
        """ default: 1x
        """
        return Decimal('1.0')

    @classmethod
    def default_date(cls):
        """ default: today
        """
        IrDate = Pool().get('ir.date')
        return IrDate.today()

# end WatchlistEntry
