# -*- coding: utf-8 -*-
# This file is part of the investment-module from m-ds for Tryton.
# The COPYRIGHT file at the top level of this repository contains the
# full copyright notices and license terms.

from trytond.tests.test_tryton import with_transaction
from trytond.pool import Pool
from trytond.transaction import Transaction
from datetime import date
from decimal import Decimal
from trytond.modules.investment.tests.test_module import InvestmentTestCase


class WatchlistTestCase(InvestmentTestCase):
    'Test watchlist module'
    module = 'investment_watchlist'

    def prep_2nd_currency(self, company):
        """ add EUR as 2nd currency
        """
        pool = Pool()
        Currency = pool.get('currency.currency')
        CurrencyRate = pool.get('currency.currency.rate')
        Company = pool.get('company.company')

        usd, = Currency.search([('name', '=', 'usd')])
        euros = Currency.search([('code', '=', 'EUR')])
        if len(euros) == 0:
            euro, = Currency.create([{
                'name': 'Euro',
                'symbol': '€',
                'code': 'EUR',
                'numeric_code': '978',
                'rounding': Decimal('0.01'),
                'digits': 2,
                }])
        else:
            euro = euros[0]

        # set company-currency to euro
        self.assertEqual(company.currency.name, 'usd')
        Company.write(*[
            [company],
            {
                'currency': euro.id,
            }])
        self.assertEqual(company.currency.name, 'Euro')

        # add rate for euro/usd @ 05/02/2022
        # EUR is base-currency
        CurrencyRate.create([{
            'date': date(2022, 5, 2),
            'currency': euro.id,
            'rate': Decimal('1.0'),
            }, {
            'date': date(2022, 5, 2),
            'currency': usd.id,
            'rate': Decimal('1.05'),
            }])

        # delete unwanted rates
        usd_1 = CurrencyRate.search([
            ('currency.id', '=', usd.id),
            ('date', '!=', date(2022, 5, 2)),
            ])
        CurrencyRate.delete(usd_1)
        return (usd, euro)

    def prep_watchlist_assets(self):
        """ create asset, fill rates
        """
        Asset = Pool().get('investment.asset')

        company = self.prep_asset_company()
        with Transaction().set_context({'company': company.id}):
            product = self.prep_asset_product(
                name='Product 1',
                description='some asset')

            asset = self.prep_asset_item(company=company, product=product)
            self.assertEqual(asset.symbol, 'usd/u')

            Asset.write(*[
                [asset],
                {
                    'rates': [('create', [{
                        'date': date(2022, 5, 1),
                        'rate': Decimal('2.5'),
                        }, {
                        'date': date(2022, 5, 2),
                        'rate': Decimal('2.8'),
                        }])],
                }])
            self.assertEqual(len(asset.rates), 2)
            self.assertEqual(asset.rate, Decimal('2.8'))
            self.assertEqual(asset.date, date(2022, 5, 2))
        return asset

    @with_transaction()
    def test_watchlist_line_amounts(self):
        """ check amounts of asset in 2nd-currency and
            watchlist-currency
        """
        pool = Pool()
        Asset = pool.get('investment.asset')
        Currency = pool.get('currency.currency')
        Watchlist = pool.get('investment_watchlist.list')
        WlEntry = pool.get('investment_watchlist.line')
        cursor = Transaction().connection.cursor()

        company = self.prep_asset_company()
        with Transaction().set_context({'company': company.id}):
            product1 = self.prep_asset_product(
                name='Product 1', description='some asset')
            product2 = self.prep_asset_product(
                name='Product 2', description='some asset 2')

            asset1 = self.prep_asset_item(company=company, product=product1)
            self.assertEqual(asset1.rec_name, 'Product 1 | - usd/u | -')

            asset2 = self.prep_asset_item(company=company, product=product2)
            self.assertEqual(asset2.rec_name, 'Product 2 | - usd/u | -')

            # switch company-currency --> euro
            # rates @ 2022-05-02, eur=1.0, usd=1.05
            self.prep_2nd_currency(company)

            # 2nd asset --> euro
            currency_euro, = Currency.search([('code', '=', 'EUR')])
            Asset.write(*[
                [asset2],
                {
                    'currency': currency_euro.id,
                }])
            self.assertEqual(asset2.rec_name, 'Product 2 | - €/u | -')

            # add rates to assets
            Asset.write(*[
                # usd
                [asset1],
                {
                    'rates': [('create', [{
                        'date': date(2022, 5, 5),
                        'rate': Decimal('5.0'),
                        }, ])],
                },
                # euro
                [asset2],
                {
                    'rates': [('create', [{
                        'date': date(2022, 5, 5),
                        'rate': Decimal('7.5'),
                        }, ])],
                }, ])

            with Transaction().set_context({
                    'qdate': date(2022, 5, 5)}):
                wl, = Watchlist.create([{
                    'name': 'List 1',
                    'currency': currency_euro.id,
                    'price_digits': 4,
                    'uom_digits': 4,
                    'lines': [('create', [{
                        'asset': asset1.id,
                        'date': date(2022, 5, 3),
                        'price': Decimal('4.5'),
                        'quantity': Decimal('2.0'),
                        'expenses': Decimal('0.5'),
                        }, {
                        'asset': asset2.id,
                        'date': date(2022, 5, 3),
                        'price': Decimal('6.5'),
                        'quantity': Decimal('2.0'),
                        'expenses': Decimal('0.5'),
                        }])],
                    }])
                self.assertEqual(wl.name, 'List 1')
                self.assertEqual(wl.currency.rec_name, 'Euro')
                self.assertEqual(wl.currency_digits, 2)
                self.assertEqual(len(wl.lines), 2)
                self.assertEqual(wl.lines[0].currency.code, 'EUR')
                self.assertEqual(wl.lines[0].asset.currency.code, 'usd')
                self.assertEqual(wl.lines[1].currency.code, 'EUR')
                self.assertEqual(wl.lines[1].asset.currency.code, 'EUR')

                tab_entry = WlEntry.__table__()
                (tab_amounts, tab2) = WlEntry.get_line_amounts_sql()
                # asset 1
                query = tab_entry.join(
                        tab_amounts,
                        condition=tab_entry.id == tab_amounts.id,
                    ).select(
                        tab_amounts.asset_date,
                        tab_amounts.obs_amount,
                        tab_amounts.obs_amount_cur,
                        tab_amounts.asset_rate,
                        tab_amounts.asset_rate_cur,
                        tab_amounts.asset_amount,
                        tab_amounts.asset_amount_cur,
                        where=tab_entry.id == wl.lines[0].id,
                    )
                cursor.execute(*query)
                result, = cursor.fetchall()
                self.assertEqual(result[0], date(2022, 5, 5))
                self.assertEqual(result[1], Decimal('9.0'))     # 2x 4.5 usd
                self.assertEqual(
                    result[2],
                    Decimal('8.5714'))  # 2x 4.5 usd --> eur
                self.assertEqual(result[3], Decimal('5.0'))     # 5.0 usd
                self.assertEqual(
                    result[4],
                    Decimal('4.7619'))  # 5.0 usd --> eur
                self.assertEqual(result[5], Decimal('10.0'))    # 2 x 5.0 usd
                self.assertEqual(
                    result[6],
                    Decimal('9.52'))    # 2 x 5.0 usd --> eur

                # asset 2
                query = tab_entry.join(
                        tab_amounts,
                        condition=tab_entry.id == tab_amounts.id,
                    ).select(
                        tab_amounts.asset_date,
                        tab_amounts.obs_amount,
                        tab_amounts.obs_amount_cur,
                        tab_amounts.asset_rate,
                        tab_amounts.asset_rate_cur,
                        tab_amounts.asset_amount,
                        tab_amounts.asset_amount_cur,
                        where=tab_entry.id == wl.lines[1].id,
                    )
                cursor.execute(*query)
                result, = cursor.fetchall()
                self.assertEqual(result[0], date(2022, 5, 5))
                self.assertEqual(result[1], Decimal('13.0'))    # 2x 6.5 eur
                self.assertEqual(
                    result[2],
                    Decimal('13.0'))    # 2x 6.5 eur --> eur
                self.assertEqual(result[3], Decimal('7.5'))     # 7.5 eur
                self.assertEqual(
                    result[4],
                    Decimal('7.5'))     # 7.5 eur --> eur
                self.assertEqual(result[5], Decimal('15.0'))    # 2 x 7.5 eur
                self.assertEqual(
                    result[6],
                    Decimal('15.0'))    # 2 x 7.5 eur --> eur

                (tab_values, tab_e) = WlEntry.get_entry_data_sql()
                tab_values.where = tab_e.id == wl.lines[0].id

                wlists = Watchlist.search([])
                self.assertEqual(len(wlists), 1)
                self.assertEqual(wlists[0].name, 'List 1')
                self.assertEqual(wlists[0].currency.rec_name, 'Euro')
                self.assertEqual(wlists[0].currency_digits, 2)
                self.assertEqual(wlists[0].price_digits, 4)
                self.assertEqual(wlists[0].uom_digits, 4)
                self.assertEqual(wlists[0].amount, Decimal('22.57'))
                self.assertEqual(wlists[0].asset_amount, Decimal('24.52'))
                self.assertEqual(wlists[0].diff_amount, Decimal('1.95'))
                self.assertEqual(wlists[0].diff_percent, Decimal('8.63'))
                self.assertEqual(len(wlists[0].lines), 2)

                self.assertEqual(
                    wlists[0].lines[0].asset.rec_name,
                    'Product 1 | 5.0000 usd/u | 05/05/2022')
                self.assertEqual(
                    wlists[0].lines[0].asset_currency.rec_name,
                    'usd')
                self.assertEqual(wlists[0].lines[0].currency.rec_name, 'Euro')
                self.assertEqual(wlists[0].lines[0].currency_digits, 2)
                self.assertEqual(wlists[0].lines[0].price_digits, 4)
                self.assertEqual(wlists[0].lines[0].uom.rec_name, 'Unit')
                self.assertEqual(wlists[0].lines[0].uom_digits, 4)
                self.assertEqual(wlists[0].lines[0].asset_digits, 4)
                self.assertEqual(wlists[0].lines[0].amount_digits, 4)
                self.assertEqual(wlists[0].lines[0].symbol, 'usd/u')
                self.assertEqual(wlists[0].lines[0].date, date(2022, 5, 3))
                self.assertEqual(wlists[0].lines[0].price, Decimal('4.5'))
                self.assertEqual(wlists[0].lines[0].quantity, Decimal('2.0'))
                self.assertEqual(wlists[0].lines[0].expenses, Decimal('0.5'))
                self.assertEqual(wlists[0].lines[0].amount, Decimal('9.0714'))
                self.assertEqual(
                    wlists[0].lines[0].asset_date,
                    date(2022, 5, 5))
                self.assertEqual(wlists[0].lines[0].asset_rate, Decimal('5.0'))
                self.assertEqual(
                    wlists[0].lines[0].asset_rate_cur,
                    Decimal('4.7619'))
                self.assertEqual(
                    wlists[0].lines[0].asset_amount,
                    Decimal('9.52'))
                self.assertEqual(
                    wlists[0].lines[0].diff_amount,
                    Decimal('0.45'))
                self.assertEqual(
                    wlists[0].lines[0].diff_percent,
                    Decimal('4.95'))

                self.assertEqual(
                    wlists[0].lines[1].asset.rec_name,
                    'Product 2 | 7.5000 €/u | 05/05/2022')
                self.assertEqual(
                    wlists[0].lines[1].asset_currency.rec_name,
                    'Euro')
                self.assertEqual(wlists[0].lines[1].currency.rec_name, 'Euro')
                self.assertEqual(wlists[0].lines[1].currency_digits, 2)
                self.assertEqual(wlists[0].lines[1].price_digits, 4)
                self.assertEqual(wlists[0].lines[1].uom.rec_name, 'Unit')
                self.assertEqual(wlists[0].lines[1].uom_digits, 4)
                self.assertEqual(wlists[0].lines[1].asset_digits, 4)
                self.assertEqual(wlists[0].lines[1].amount_digits, 4)
                self.assertEqual(wlists[0].lines[1].symbol, '€/u')
                self.assertEqual(wlists[0].lines[1].date, date(2022, 5, 3))
                self.assertEqual(wlists[0].lines[1].price, Decimal('6.5'))
                self.assertEqual(wlists[0].lines[1].quantity, Decimal('2.0'))
                self.assertEqual(wlists[0].lines[1].expenses, Decimal('0.5'))
                self.assertEqual(wlists[0].lines[1].amount, Decimal('13.5'))
                self.assertEqual(
                    wlists[0].lines[1].asset_date,
                    date(2022, 5, 5))
                self.assertEqual(wlists[0].lines[1].asset_rate, Decimal('7.5'))
                self.assertEqual(
                    wlists[0].lines[1].asset_rate_cur,
                    Decimal('7.5'))
                self.assertEqual(
                    wlists[0].lines[1].asset_amount,
                    Decimal('15.0'))
                self.assertEqual(wlists[0].lines[1].diff_amount, Decimal('1.5'))
                self.assertEqual(
                    wlists[0].lines[1].diff_percent,
                    Decimal('11.11'))

    @with_transaction()
    def test_watchlist_twice_amounts(self):
        """ add 2x watchlist, 5x assets, 2x asset in both WL
        """
        pool = Pool()
        Asset = pool.get('investment.asset')
        Currency = pool.get('currency.currency')
        Watchlist = pool.get('investment_watchlist.list')

        company = self.prep_asset_company()
        with Transaction().set_context({'company': company.id}):
            asset1 = self.prep_asset_item(
                company=company,
                product=self.prep_asset_product(name='Product 1'))
            asset2 = self.prep_asset_item(
                company=company,
                product=self.prep_asset_product(name='Product 2'))
            asset3 = self.prep_asset_item(
                company=company,
                product=self.prep_asset_product(name='Product 3'))
            asset4 = self.prep_asset_item(
                company=company,
                product=self.prep_asset_product(name='Product 4'))
            asset5 = self.prep_asset_item(
                company=company,
                product=self.prep_asset_product(name='Product 5'))

            self.assertEqual(asset1.rec_name, 'Product 1 | - usd/u | -')
            self.assertEqual(asset2.rec_name, 'Product 2 | - usd/u | -')
            self.assertEqual(asset3.rec_name, 'Product 3 | - usd/u | -')
            self.assertEqual(asset4.rec_name, 'Product 4 | - usd/u | -')
            self.assertEqual(asset5.rec_name, 'Product 5 | - usd/u | -')

            # switch company-currency --> euro
            # rates @ 2022-05-02, eur=1.0, usd=1.05
            self.prep_2nd_currency(company)

            # assets --> euro
            currency_euro, = Currency.search([('code', '=', 'EUR')])
            currency_usd, = Currency.search([('code', '=', 'usd')])
            Asset.write(*[
                [asset2, asset3, asset5],
                {
                    'currency': currency_euro.id,
                }])
            self.assertEqual(asset1.rec_name, 'Product 1 | - usd/u | -')
            self.assertEqual(asset2.rec_name, 'Product 2 | - €/u | -')
            self.assertEqual(asset3.rec_name, 'Product 3 | - €/u | -')
            self.assertEqual(asset4.rec_name, 'Product 4 | - usd/u | -')
            self.assertEqual(asset5.rec_name, 'Product 5 | - €/u | -')

            # add rates to assets
            Asset.write(*[
                # usd
                [asset1, asset2, asset4],
                {
                    'rates': [('create', [{
                        'date': date(2022, 5, 5),
                        'rate': Decimal('5.0'),
                        }, ])],
                },
                # euro
                [asset3, asset5],
                {
                    'rates': [('create', [{
                        'date': date(2022, 5, 4),
                        'rate': Decimal('7.5'),
                        }, ])],
                }, ])
            self.assertEqual(
                asset1.rec_name,
                'Product 1 | 5.0000 usd/u | 05/05/2022')
            self.assertEqual(asset2.rec_name, 'Product 2 | 5.0000 €/u | 05/05/2022')
            self.assertEqual(asset3.rec_name, 'Product 3 | 7.5000 €/u | 05/04/2022')
            self.assertEqual(
                asset4.rec_name,
                'Product 4 | 5.0000 usd/u | 05/05/2022')
            self.assertEqual(asset5.rec_name, 'Product 5 | 7.5000 €/u | 05/04/2022')

            with Transaction().set_context({
                    'qdate': date(2022, 5, 5)}):
                wlists = Watchlist.create([{
                    'name': 'List 1',
                    'currency': currency_euro.id,
                    'price_digits': 2,
                    'uom_digits': 3,
                    'lines': [('create', [{
                        'asset': asset1.id,
                        'date': date(2022, 5, 3),
                        'price': Decimal('4.5'),
                        'quantity': Decimal('2.0'),
                        'expenses': Decimal('0.5'),
                        }, {
                        'asset': asset2.id,
                        'date': date(2022, 5, 3),
                        'price': Decimal('6.5'),
                        'quantity': Decimal('2.0'),
                        'expenses': Decimal('0.5'),
                        }, {
                        'asset': asset4.id,
                        'date': date(2022, 5, 3),
                        'price': Decimal('5.1'),
                        'quantity': Decimal('1.5'),
                        'expenses': Decimal('0.3'),
                        }])],
                    }, {
                    'name': 'List 2',
                    'currency': currency_usd.id,
                    'price_digits': 3,
                    'uom_digits': 2,
                    'lines': [('create', [{
                        'asset': asset2.id,
                        'date': date(2022, 5, 3),
                        'price': Decimal('4.75'),
                        'quantity': Decimal('2.0'),
                        'expenses': Decimal('0.5'),
                        }, {
                        'asset': asset3.id,
                        'date': date(2022, 5, 3),
                        'price': Decimal('6.5'),
                        'quantity': Decimal('2.0'),
                        'expenses': Decimal('0.5'),
                        }, {
                        'asset': asset4.id,
                        'date': date(2022, 5, 3),
                        'price': Decimal('5.5'),
                        'quantity': Decimal('1.0'),
                        'expenses': Decimal('0.25'),
                        }, {
                        'asset': asset5.id,
                        'date': date(2022, 5, 3),
                        'price': Decimal('6.8'),
                        'quantity': Decimal('1.0'),
                        'expenses': Decimal('0.25'),
                        }])],
                    }])
                self.assertEqual(len(wlists), 2)

                self.assertEqual(wlists[0].name, 'List 1')
                self.assertEqual(wlists[0].currency.rec_name, 'Euro')
                self.assertEqual(wlists[0].currency_digits, 2)
                self.assertEqual(wlists[0].price_digits, 2)
                self.assertEqual(wlists[0].uom_digits, 3)
                self.assertEqual(wlists[0].amount, Decimal('30.16'))
                self.assertEqual(wlists[0].asset_amount, Decimal('26.66'))
                self.assertEqual(wlists[0].diff_amount, Decimal('-3.50'))
                self.assertEqual(wlists[0].diff_percent, Decimal('-11.6'))

                self.assertEqual(len(wlists[0].lines), 3)

                # list 1 / asset 1
                self.assertEqual(
                    wlists[0].lines[0].asset.rec_name,
                    'Product 1 | 5.0000 usd/u | 05/05/2022')
                self.assertEqual(wlists[0].lines[0].asset_currency.rec_name, 'usd')
                self.assertEqual(wlists[0].lines[0].currency.rec_name, 'Euro')
                self.assertEqual(wlists[0].lines[0].currency_digits, 2)
                self.assertEqual(wlists[0].lines[0].price_digits, 2)
                self.assertEqual(wlists[0].lines[0].uom.rec_name, 'Unit')
                self.assertEqual(wlists[0].lines[0].uom_digits, 3)
                self.assertEqual(wlists[0].lines[0].asset_digits, 4)
                self.assertEqual(wlists[0].lines[0].amount_digits, 3)
                self.assertEqual(wlists[0].lines[0].symbol, 'usd/u')
                self.assertEqual(wlists[0].lines[0].date, date(2022, 5, 3))
                self.assertEqual(wlists[0].lines[0].price, Decimal('4.5'))
                self.assertEqual(wlists[0].lines[0].quantity, Decimal('2.0'))
                self.assertEqual(wlists[0].lines[0].expenses, Decimal('0.5'))
                self.assertEqual(wlists[0].lines[0].amount, Decimal('9.071'))
                self.assertEqual(wlists[0].lines[0].asset_date, date(2022, 5, 5))
                self.assertEqual(wlists[0].lines[0].asset_rate, Decimal('5.0'))
                self.assertEqual(
                    wlists[0].lines[0].asset_rate_cur,
                    Decimal('4.7619'))
                self.assertEqual(wlists[0].lines[0].asset_amount, Decimal('9.52'))
                self.assertEqual(wlists[0].lines[0].diff_amount, Decimal('0.45'))
                self.assertEqual(wlists[0].lines[0].diff_percent, Decimal('4.95'))
                self.assertEqual(
                    wlists[0].lines[0].rec_name,
                    'Product 1 | 9.52 € (4.95 %, 0.45 €) | 05/05/2022')

                # list 1 / asset 2
                self.assertEqual(
                    wlists[0].lines[1].asset.rec_name,
                    'Product 2 | 5.0000 €/u | 05/05/2022')
                self.assertEqual(wlists[0].lines[1].asset_currency.rec_name, 'Euro')
                self.assertEqual(wlists[0].lines[1].currency.rec_name, 'Euro')
                self.assertEqual(wlists[0].lines[1].currency_digits, 2)
                self.assertEqual(wlists[0].lines[1].price_digits, 2)
                self.assertEqual(wlists[0].lines[1].uom.rec_name, 'Unit')
                self.assertEqual(wlists[0].lines[1].uom_digits, 3)
                self.assertEqual(wlists[0].lines[1].asset_digits, 4)
                self.assertEqual(wlists[0].lines[1].amount_digits, 3)
                self.assertEqual(wlists[0].lines[1].symbol, '€/u')
                self.assertEqual(wlists[0].lines[1].date, date(2022, 5, 3))
                self.assertEqual(wlists[0].lines[1].price, Decimal('6.5'))
                self.assertEqual(wlists[0].lines[1].quantity, Decimal('2.0'))
                self.assertEqual(wlists[0].lines[1].expenses, Decimal('0.5'))
                self.assertEqual(wlists[0].lines[1].amount, Decimal('13.5'))
                self.assertEqual(wlists[0].lines[1].asset_date, date(2022, 5, 5))
                self.assertEqual(wlists[0].lines[1].asset_rate, Decimal('5.0'))
                self.assertEqual(wlists[0].lines[1].asset_rate_cur, Decimal('5.0'))
                self.assertEqual(wlists[0].lines[1].asset_amount, Decimal('10.0'))
                self.assertEqual(wlists[0].lines[1].diff_amount, Decimal('-3.5'))
                self.assertEqual(wlists[0].lines[1].diff_percent, Decimal('-25.93'))
                self.assertEqual(
                    wlists[0].lines[1].rec_name,
                    'Product 2 | 10.00 € (-25.93 %, -3.50 €) | 05/05/2022')

                # list 1 / asset 3
                self.assertEqual(
                    wlists[0].lines[2].asset.rec_name,
                    'Product 4 | 5.0000 usd/u | 05/05/2022')
                self.assertEqual(wlists[0].lines[2].asset_currency.rec_name, 'usd')
                self.assertEqual(wlists[0].lines[2].currency.rec_name, 'Euro')
                self.assertEqual(wlists[0].lines[2].currency_digits, 2)
                self.assertEqual(wlists[0].lines[2].price_digits, 2)
                self.assertEqual(wlists[0].lines[2].uom.rec_name, 'Unit')
                self.assertEqual(wlists[0].lines[2].uom_digits, 3)
                self.assertEqual(wlists[0].lines[2].asset_digits, 4)
                self.assertEqual(wlists[0].lines[2].amount_digits, 3)
                self.assertEqual(wlists[0].lines[2].symbol, 'usd/u')
                self.assertEqual(wlists[0].lines[2].date, date(2022, 5, 3))
                self.assertEqual(wlists[0].lines[2].price, Decimal('5.1'))
                self.assertEqual(wlists[0].lines[2].quantity, Decimal('1.5'))
                self.assertEqual(wlists[0].lines[2].expenses, Decimal('0.3'))
                self.assertEqual(wlists[0].lines[2].amount, Decimal('7.586'))
                self.assertEqual(wlists[0].lines[2].asset_date, date(2022, 5, 5))
                self.assertEqual(wlists[0].lines[2].asset_rate, Decimal('5.0'))
                self.assertEqual(
                    wlists[0].lines[2].asset_rate_cur,
                    Decimal('4.7619'))
                self.assertEqual(wlists[0].lines[2].asset_amount, Decimal('7.14'))
                self.assertEqual(wlists[0].lines[2].diff_amount, Decimal('-0.45'))
                self.assertEqual(wlists[0].lines[2].diff_percent, Decimal('-5.88'))

                # list 2
                self.assertEqual(wlists[1].name, 'List 2')
                self.assertEqual(wlists[1].currency.rec_name, 'usd')
                self.assertEqual(wlists[1].currency_digits, 2)
                self.assertEqual(wlists[1].price_digits, 3)
                self.assertEqual(wlists[1].uom_digits, 2)
                self.assertEqual(wlists[1].amount, Decimal('37.77'))
                self.assertEqual(wlists[1].asset_amount, Decimal('39.13'))
                self.assertEqual(wlists[1].diff_amount, Decimal('1.36'))
                self.assertEqual(wlists[1].diff_percent, Decimal('3.6'))

                self.assertEqual(len(wlists[1].lines), 4)

                # list 2 / asset 1
                self.assertEqual(
                    wlists[1].lines[0].asset.rec_name,
                    'Product 2 | 5.0000 €/u | 05/05/2022')
                self.assertEqual(wlists[1].lines[0].asset_currency.rec_name, 'Euro')
                self.assertEqual(wlists[1].lines[0].currency.rec_name, 'usd')
                self.assertEqual(wlists[1].lines[0].currency_digits, 2)
                self.assertEqual(wlists[1].lines[0].price_digits, 3)
                self.assertEqual(wlists[1].lines[0].uom.rec_name, 'Unit')
                self.assertEqual(wlists[1].lines[0].uom_digits, 2)
                self.assertEqual(wlists[1].lines[0].asset_digits, 4)
                self.assertEqual(wlists[1].lines[0].amount_digits, 2)
                self.assertEqual(wlists[1].lines[0].symbol, '€/u')
                self.assertEqual(wlists[1].lines[0].date, date(2022, 5, 3))
                self.assertEqual(wlists[1].lines[0].price, Decimal('4.75'))
                self.assertEqual(wlists[1].lines[0].quantity, Decimal('2.0'))
                self.assertEqual(wlists[1].lines[0].expenses, Decimal('0.5'))
                self.assertEqual(wlists[1].lines[0].amount, Decimal('10.48'))
                self.assertEqual(wlists[1].lines[0].asset_date, date(2022, 5, 5))
                self.assertEqual(wlists[1].lines[0].asset_rate, Decimal('5.0'))
                self.assertEqual(wlists[1].lines[0].asset_rate_cur, Decimal('5.25'))
                self.assertEqual(wlists[1].lines[0].asset_amount, Decimal('10.5'))
                self.assertEqual(wlists[1].lines[0].diff_amount, Decimal('0.02'))
                self.assertEqual(wlists[1].lines[0].diff_percent, Decimal('0.19'))

                # list 1 / asset 2
                self.assertEqual(
                    wlists[1].lines[1].asset.rec_name,
                    'Product 4 | 5.0000 usd/u | 05/05/2022')
                self.assertEqual(wlists[1].lines[1].asset_currency.rec_name, 'usd')
                self.assertEqual(wlists[1].lines[1].currency.rec_name, 'usd')
                self.assertEqual(wlists[1].lines[1].currency_digits, 2)
                self.assertEqual(wlists[1].lines[1].price_digits, 3)
                self.assertEqual(wlists[1].lines[1].uom.rec_name, 'Unit')
                self.assertEqual(wlists[1].lines[1].uom_digits, 2)
                self.assertEqual(wlists[1].lines[1].asset_digits, 4)
                self.assertEqual(wlists[1].lines[1].amount_digits, 2)
                self.assertEqual(wlists[1].lines[1].symbol, 'usd/u')
                self.assertEqual(wlists[1].lines[1].date, date(2022, 5, 3))
                self.assertEqual(wlists[1].lines[1].price, Decimal('5.5'))
                self.assertEqual(wlists[1].lines[1].quantity, Decimal('1.0'))
                self.assertEqual(wlists[1].lines[1].expenses, Decimal('0.25'))
                self.assertEqual(wlists[1].lines[1].amount, Decimal('5.75'))
                self.assertEqual(wlists[1].lines[1].asset_date, date(2022, 5, 5))
                self.assertEqual(wlists[1].lines[1].asset_rate, Decimal('5.0'))
                self.assertEqual(wlists[1].lines[1].asset_rate_cur, Decimal('5.0'))
                self.assertEqual(wlists[1].lines[1].asset_amount, Decimal('5.0'))
                self.assertEqual(wlists[1].lines[1].diff_amount, Decimal('-0.75'))
                self.assertEqual(wlists[1].lines[1].diff_percent, Decimal('-13.04'))

                # list 2 / asset 3
                self.assertEqual(
                    wlists[1].lines[2].asset.rec_name,
                    'Product 3 | 7.5000 €/u | 05/04/2022')
                self.assertEqual(wlists[1].lines[2].asset_currency.rec_name, 'Euro')
                self.assertEqual(wlists[1].lines[2].currency.rec_name, 'usd')
                self.assertEqual(wlists[1].lines[2].currency_digits, 2)
                self.assertEqual(wlists[1].lines[2].price_digits, 3)
                self.assertEqual(wlists[1].lines[2].uom.rec_name, 'Unit')
                self.assertEqual(wlists[1].lines[2].uom_digits, 2)
                self.assertEqual(wlists[1].lines[2].asset_digits, 4)
                self.assertEqual(wlists[1].lines[2].amount_digits, 2)
                self.assertEqual(wlists[1].lines[2].symbol, '€/u')
                self.assertEqual(wlists[1].lines[2].date, date(2022, 5, 3))
                self.assertEqual(wlists[1].lines[2].price, Decimal('6.5'))
                self.assertEqual(wlists[1].lines[2].quantity, Decimal('2.0'))
                self.assertEqual(wlists[1].lines[2].expenses, Decimal('0.5'))
                self.assertEqual(wlists[1].lines[2].amount, Decimal('14.15'))
                self.assertEqual(wlists[1].lines[2].asset_date, date(2022, 5, 4))
                self.assertEqual(wlists[1].lines[2].asset_rate, Decimal('7.5'))
                self.assertEqual(
                    wlists[1].lines[2].asset_rate_cur,
                    Decimal('7.875'))
                self.assertEqual(wlists[1].lines[2].asset_amount, Decimal('15.75'))
                self.assertEqual(wlists[1].lines[2].diff_amount, Decimal('1.6'))
                self.assertEqual(wlists[1].lines[2].diff_percent, Decimal('11.31'))

                # list 2 / asset 4
                self.assertEqual(
                    wlists[1].lines[3].asset.rec_name,
                    'Product 5 | 7.5000 €/u | 05/04/2022')
                self.assertEqual(wlists[1].lines[3].asset_currency.rec_name, 'Euro')
                self.assertEqual(wlists[1].lines[3].currency.rec_name, 'usd')
                self.assertEqual(wlists[1].lines[3].currency_digits, 2)
                self.assertEqual(wlists[1].lines[3].price_digits, 3)
                self.assertEqual(wlists[1].lines[3].uom.rec_name, 'Unit')
                self.assertEqual(wlists[1].lines[3].uom_digits, 2)
                self.assertEqual(wlists[1].lines[3].asset_digits, 4)
                self.assertEqual(wlists[1].lines[3].amount_digits, 2)
                self.assertEqual(wlists[1].lines[3].symbol, '€/u')
                self.assertEqual(wlists[1].lines[3].date, date(2022, 5, 3))
                self.assertEqual(wlists[1].lines[3].price, Decimal('6.8'))
                self.assertEqual(wlists[1].lines[3].quantity, Decimal('1.0'))
                self.assertEqual(wlists[1].lines[3].expenses, Decimal('0.25'))
                self.assertEqual(wlists[1].lines[3].amount, Decimal('7.39'))
                self.assertEqual(wlists[1].lines[3].asset_date, date(2022, 5, 4))
                self.assertEqual(wlists[1].lines[3].asset_rate, Decimal('7.5'))
                self.assertEqual(
                    wlists[1].lines[3].asset_rate_cur,
                    Decimal('7.875'))
                self.assertEqual(wlists[1].lines[3].asset_amount, Decimal('7.88'))
                self.assertEqual(wlists[1].lines[3].diff_amount, Decimal('0.49'))
                self.assertEqual(wlists[1].lines[3].diff_percent, Decimal('6.63'))

    @with_transaction()
    def test_watchlist_currency_rate(self):
        """ check currency-rate
        """
        pool = Pool()
        WlEntry = pool.get('investment_watchlist.line')
        Currency = pool.get('currency.currency')
        cursor = Transaction().connection.cursor()

        Currency.create([{
            'name': 'Euro',
            'symbol': '€',
            'code': 'EUR',
            'numeric_code': '978',
            'rounding': Decimal('0.01'),
            'digits': 2,
            'rates': [('create', [{
                'date': date(2022, 4, 30),
                'rate': Decimal('0.0'),
                }, {
                'date': date(2022, 5, 1),
                'rate': Decimal('1.1'),
                }, {
                'date': date(2022, 5, 2),
                'rate': Decimal('1.2'),
                }, {
                'date': date(2022, 5, 3),
                'rate': Decimal('1.3'),
                }])],
            }])

        with Transaction().set_context({
                'qdate': date(2022, 4, 29)}):
            tab_cur = WlEntry.get_currency_rate_sql()
            cursor.execute(*tab_cur)
            l1 = cursor.fetchall()
            self.assertEqual(len(l1), 1)
            (id1, digits, value) = l1[0]
            # no rate @ 2022-04-29 or before, set rate to 1.0
            self.assertEqual(value, Decimal('1.0'))
            self.assertEqual(digits, 2)

        with Transaction().set_context({
                'qdate': date(2022, 4, 30)}):
            tab_cur = WlEntry.get_currency_rate_sql()
            cursor.execute(*tab_cur)
            l1 = cursor.fetchall()
            self.assertEqual(len(l1), 1)
            (id1, digits, value) = l1[0]
            # rate @ 2022-04-30: 0.0, causes divide by zero
            # updated to 1.0
            self.assertEqual(value, Decimal('1.0'))
            self.assertEqual(digits, 2)

        with Transaction().set_context({
                'qdate': date(2022, 5, 1)}):
            tab_cur = WlEntry.get_currency_rate_sql()
            cursor.execute(*tab_cur)
            l1 = cursor.fetchall()
            self.assertEqual(len(l1), 1)
            (id1, digits, value) = l1[0]
            self.assertEqual(value, Decimal('1.1'))
            self.assertEqual(digits, 2)

        with Transaction().set_context({
                'qdate': date(2022, 5, 2)}):
            tab_cur = WlEntry.get_currency_rate_sql()
            cursor.execute(*tab_cur)
            l1 = cursor.fetchall()
            self.assertEqual(len(l1), 1)
            (id1, digits, value) = l1[0]
            self.assertEqual(value, Decimal('1.2'))
            self.assertEqual(digits, 2)

        with Transaction().set_context({
                'qdate': date(2022, 5, 5)}):
            tab_cur = WlEntry.get_currency_rate_sql()
            cursor.execute(*tab_cur)
            l1 = cursor.fetchall()
            self.assertEqual(len(l1), 1)
            (id1, digits, value) = l1[0]
            self.assertEqual(value, Decimal('1.3'))
            self.assertEqual(digits, 2)

    @with_transaction()
    def test_watchlist_create(self):
        """ create asset, fill rates
        """
        pool = Pool()
        WatchList = pool.get('investment_watchlist.list')
        Rate = pool.get('investment.rate')

        asset = self.prep_watchlist_assets()

        with Transaction().set_context({
                'company': asset.company.id,
                'qdate': date(2022, 5, 5)}):
            watchl, = WatchList.create([{
                'name': 'List 1',
                'lines': [('create', [{
                    'asset': asset.id,
                    'date': date(2022, 4, 1),
                    'price': Decimal('2.1'),
                    'quantity': Decimal('25.0'),
                    'expenses': Decimal('4.20'),
                    }])],
                }])

            self.assertEqual(len(watchl.lines), 1)
            self.assertEqual(watchl.lines[0].amount, Decimal('56.7'))
            self.assertEqual(watchl.lines[0].date, date(2022, 4, 1))
            self.assertEqual(watchl.lines[0].price, Decimal('2.1'))
            self.assertEqual(watchl.lines[0].quantity, Decimal('25.0'))
            self.assertEqual(watchl.lines[0].expenses, Decimal('4.2'))

            self.assertEqual(watchl.lines[0].asset_date, date(2022, 5, 2))
            self.assertEqual(watchl.lines[0].asset_rate, Decimal('2.8'))
            self.assertEqual(watchl.lines[0].asset_amount, Decimal('70.0'))

            self.assertEqual(watchl.amount, Decimal('56.7'))
            self.assertEqual(watchl.asset_amount, Decimal('70.0'))
            self.assertEqual(watchl.diff_amount, Decimal('13.3'))
            self.assertEqual(watchl.diff_percent, Decimal('23.46'))

            # 70.0 - 48.3
            self.assertEqual(watchl.lines[0].diff_amount, Decimal('13.3'))
            self.assertEqual(watchl.lines[0].diff_percent, Decimal('23.46'))

            # check: prevention against divide by zero
            WatchList.write(*[
                [watchl],
                {
                    'lines': [('write', [watchl.lines[0]], {
                        'quantity': Decimal('0.0'),
                        })],
                }])
            self.assertEqual(watchl.lines[0].quantity, Decimal('0'))
            self.assertEqual(watchl.lines[0].asset_rate, Decimal('2.8'))
            self.assertEqual(watchl.lines[0].asset_amount, Decimal('0.0'))
            self.assertEqual(watchl.lines[0].diff_amount, Decimal('0.0'))
            self.assertEqual(watchl.lines[0].diff_percent, Decimal('0.0'))

            # prepare for specific percents
            WatchList.write(*[
                [watchl],
                {
                    'lines': [('write', [watchl.lines[0]], {
                        'quantity': Decimal('10.0'),
                        'price': Decimal('5.0'),
                        'expenses': Decimal('0.0'),
                        })],
                }])

            self.assertEqual(asset.rates[0].date, date(2022, 5, 2))
            self.assertEqual(asset.rates[0].rate, Decimal('2.8'))

            # 50 --> 75
            Rate.write(*[[asset.rates[0]], {'rate': Decimal('7.5')}])
            self.assertEqual(watchl.lines[0].asset_amount, Decimal('75.0'))
            self.assertEqual(watchl.lines[0].diff_amount, Decimal('25.0'))
            self.assertEqual(watchl.lines[0].diff_percent, Decimal('50.0'))
            self.assertEqual(watchl.amount, Decimal('50.0'))
            self.assertEqual(watchl.asset_amount, Decimal('75.0'))
            self.assertEqual(watchl.diff_amount, Decimal('25.0'))
            self.assertEqual(watchl.diff_percent, Decimal('50.0'))

            # 50 --> 100
            Rate.write(*[[asset.rates[0]], {'rate': Decimal('10.0')}])
            self.assertEqual(watchl.lines[0].asset_amount, Decimal('100.0'))
            self.assertEqual(watchl.lines[0].diff_amount, Decimal('50.0'))
            self.assertEqual(watchl.lines[0].diff_percent, Decimal('100.0'))
            self.assertEqual(watchl.amount, Decimal('50.0'))
            self.assertEqual(watchl.asset_amount, Decimal('100.0'))
            self.assertEqual(watchl.diff_amount, Decimal('50.0'))
            self.assertEqual(watchl.diff_percent, Decimal('100.0'))

            # 50 --> 150
            Rate.write(*[[asset.rates[0]], {'rate': Decimal('15.0')}])
            self.assertEqual(watchl.lines[0].asset_amount, Decimal('150.0'))
            self.assertEqual(watchl.lines[0].diff_amount, Decimal('100.0'))
            self.assertEqual(watchl.lines[0].diff_percent, Decimal('200.0'))
            self.assertEqual(watchl.amount, Decimal('50.0'))
            self.assertEqual(watchl.asset_amount, Decimal('150.0'))
            self.assertEqual(watchl.diff_amount, Decimal('100.0'))
            self.assertEqual(watchl.diff_percent, Decimal('200.0'))

            # 50 --> 200
            Rate.write(*[[asset.rates[0]], {'rate': Decimal('20.0')}])
            self.assertEqual(watchl.lines[0].asset_amount, Decimal('200.0'))
            self.assertEqual(watchl.lines[0].diff_amount, Decimal('150.0'))
            self.assertEqual(watchl.lines[0].diff_percent, Decimal('300.0'))
            self.assertEqual(watchl.amount, Decimal('50.0'))
            self.assertEqual(watchl.asset_amount, Decimal('200.0'))
            self.assertEqual(watchl.diff_amount, Decimal('150.0'))
            self.assertEqual(watchl.diff_percent, Decimal('300.0'))

            # 50 --> 25
            Rate.write(*[[asset.rates[0]], {'rate': Decimal('2.5')}])
            self.assertEqual(watchl.lines[0].asset_amount, Decimal('25.0'))
            self.assertEqual(watchl.lines[0].diff_amount, Decimal('-25.0'))
            self.assertEqual(watchl.lines[0].diff_percent, Decimal('-50.0'))

            # 50 --> 10
            Rate.write(*[[asset.rates[0]], {'rate': Decimal('1.0')}])
            self.assertEqual(watchl.lines[0].asset_amount, Decimal('10.0'))
            self.assertEqual(watchl.lines[0].diff_amount, Decimal('-40.0'))
            self.assertEqual(watchl.lines[0].diff_percent, Decimal('-80.0'))

            # 50 --> 0
            Rate.write(*[[asset.rates[0]], {'rate': Decimal('0.0')}])
            self.assertEqual(watchl.lines[0].asset_amount, Decimal('0.0'))
            self.assertEqual(watchl.lines[0].diff_amount, Decimal('-50.0'))
            self.assertEqual(watchl.lines[0].diff_percent, Decimal('-100.0'))

# end WatchlistTestCase


del InvestmentTestCase
