# -*- coding: utf-8 -*-
# This file is part of the investment-module from m-ds for Tryton.
# The COPYRIGHT file at the top level of this repository contains the
# full copyright notices and license terms.

from trytond.model import ModelView, ModelSQL, fields, Index
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.pyson import Eval, If
from sql import Literal
from sql.aggregate import Sum
from sql.functions import Round
from sql.conditionals import NullIf, Coalesce


class Watchlist(ModelSQL, ModelView):
    'Watchlist'
    __name__ = 'investment_watchlist.list'

    name = fields.Char(string='Name', required=True)
    company = fields.Many2One(
        string='Company', model_name='company.company',
        required=True, ondelete="RESTRICT")
    lines = fields.One2Many(
        string='Lines', field='watchlist',
        model_name='investment_watchlist.line')
    currency = fields.Many2One(
        string='Currency', required=True,
        model_name='currency.currency', ondelete='RESTRICT')
    currency_digits = fields.Function(fields.Integer(
        string='Currency Digits',
        readonly=True), 'on_change_with_currency_digits')
    price_digits = fields.Integer(
        string='Price Digits', required=True,
        domain=[('price_digits', '>=', 0), ('price_digits', '<=', 6)])
    uom_digits = fields.Integer(
        string='UOM Digits', required=True,
        domain=[('uom_digits', '>=', 0), ('uom_digits', '<=', 6)])

    amount = fields.Function(fields.Numeric(
        string='Purchase price',
        readonly=True, digits=(16, Eval('currency_digits', 2)),
        depends=['currency_digits']), 'get_watchlist_data')
    asset_amount = fields.Function(fields.Numeric(
        string='act.Value', help='Current value of the aquisition.',
        readonly=True, depends=['currency_digits'],
        digits=(16, Eval('currency_digits', 4))), 'get_watchlist_data')
    diff_amount = fields.Function(fields.Numeric(
        string='Difference',
        help='Difference between acquisition value and current value',
        readonly=True, digits=(16, Eval('currency_digits', 2)),
        depends=['currency_digits']), 'get_watchlist_data')
    diff_percent = fields.Function(fields.Numeric(
        string='Percent', help='percentage performance since acquisition',
        readonly=True, digits=(16, Eval('currency_digits', 2)),
        depends=['currency_digits']), 'get_watchlist_data')

    @classmethod
    def __setup__(cls):
        super(Watchlist, cls).__setup__()
        cls._order.insert(0, ('name', 'ASC'))
        t = cls.__table__()
        cls._sql_indexes.update({
            Index(
                t,
                (t.currency, Index.Equality())),
            Index(
                t,
                (t.company, Index.Equality())),
            })

    @classmethod
    def view_attributes(cls):
        return super(Watchlist, cls).view_attributes() + [
            ('/tree', 'visual',
                If(Eval('diff_percent', 0) < 0, 'danger',
                    If(Eval('diff_percent', 0) > 0, 'success', ''))),
            ]

    @staticmethod
    def default_company():
        return Transaction().context.get('company') or None

    @classmethod
    def default_currency(cls):
        """ currency of company
        """
        Company = Pool().get('company.company')

        company = cls.default_company()
        if company:
            company = Company(company)
            if company.currency:
                return company.currency.id

    @classmethod
    def default_price_digits(cls):
        """ default: 4
        """
        return 4

    @classmethod
    def default_uom_digits(cls):
        """ default: 4
        """
        return 4

    @fields.depends('currency')
    def on_change_with_currency_digits(self, name=None):
        """ get digits
        """
        if self.currency:
            return self.currency.digits

    @classmethod
    def get_watchlist_data_sql(cls):
        """ sql for watchlist
        """
        pool = Pool()
        Line = pool.get('investment_watchlist.line')
        Currency = pool.get('currency.currency')
        (tab_line, tab2) = Line.get_line_amounts_sql()
        tab_cur = Currency.__table__()
        tab_list = cls.__table__()

        query = tab_list.join(
                tab_line,
                condition=tab_line.watchlist == tab_list.id,
            ).join(
                tab_cur,
                condition=tab_cur.id == tab_list.currency,
            ).select(
                tab_list.id,
                Round(
                    Sum(tab_line.obs_amount_cur + tab_line.expenses),
                    tab_cur.digits).as_('amount'),
                Round(
                    Sum(tab_line.asset_amount_cur),
                    tab_cur.digits).as_('asset_amount'),
                Round(
                    Sum(tab_line.asset_amount_cur -
                        (tab_line.obs_amount_cur + tab_line.expenses)),
                    tab_cur.digits).as_('diff_amount'),
                Round(
                    Literal(100.0) * Sum(tab_line.asset_amount_cur) /
                    Coalesce(
                        NullIf(
                            Sum(tab_line.obs_amount_cur + tab_line.expenses),
                            0.0),
                        1.0) - Literal(100.0),
                    tab_cur.digits).as_('diff_percent'),
                group_by=[tab_list.id, tab_cur.digits])
        return (query, tab_list)

    @classmethod
    def get_watchlist_data(cls, lists, names):
        """ get values of list
        """
        cursor = Transaction().connection.cursor()

        result = {x: {y.id: None for y in lists} for x in names}
        if lists:
            (query, tab_list) = cls.get_watchlist_data_sql()
            query.where = tab_list.id.in_([x.id for x in lists])

            cursor.execute(*query)
            records = cursor.fetchall()

            for record in records:
                values = {
                    'amount': record[1],
                    'asset_amount': record[2],
                    'diff_amount': record[3],
                    'diff_percent': record[4],
                    }

                for name in names:
                    result[name][record[0]] = values[name]
        return result

# end Watchlist
